/*
   I, David Oberhollenzer, author of this file hereby place the contents of
   this file into the public domain. Please feel free to use this file in any
   way you wish.
 */

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <string.h>
#include <stdlib.h>


#define BUTTON_HOVER 0x01
#define BUTTON_CLICKED 0x02


struct button {
	int x, y;
	unsigned int width, height;
	int textx, texty;
	int flags;
	const char *text;
};

struct messagebox {
	Display *dpy;
	Window wnd;
	struct button ok;
	int black, white;
	Atom wmDelete;
	GC gc;
};

static const char *wmDeleteWindow = "WM_DELETE_WINDOW";

static int is_point_inside(struct button *b, int px, int py)
{
	return px >= b->x && px <= (b->x + (int)b->width - 1) &&
		py >= b->y && py <= (b->y + (int)b->height - 1);
}

static void messagebox_draw_button(struct messagebox *mb, struct button *b)
{
	int offset = (b->flags & BUTTON_CLICKED) ? 1 : 0;

	if (b->flags & BUTTON_HOVER) {
		XFillRectangle(mb->dpy, mb->wnd, mb->gc,
			       b->x + offset, b->y + offset,
			       b->width, b->height);
		XSetForeground(mb->dpy, mb->gc, mb->black);
		XSetBackground(mb->dpy, mb->gc, mb->white);
	} else {
		XSetForeground(mb->dpy, mb->gc, mb->white);
		XSetBackground(mb->dpy, mb->gc, mb->black);
		XDrawRectangle(mb->dpy, mb->wnd, mb->gc, b->x, b->y,
			       b->width, b->height);
	}

	XDrawString(mb->dpy, mb->wnd, mb->gc,
		    b->textx + offset, b->texty + offset,
		    b->text, strlen(b->text));
	XSetForeground(mb->dpy, mb->gc, mb->white);
	XSetBackground(mb->dpy, mb->gc, mb->black);
}

static int messagebox_init(struct messagebox *mb, const char *title)
{
	if (!(mb->dpy = XOpenDisplay(0)))
		return 0;

	mb->black = BlackPixel(mb->dpy, DefaultScreen(mb->dpy));
	mb->white = WhitePixel(mb->dpy, DefaultScreen(mb->dpy));
	mb->wmDelete = XInternAtom(mb->dpy, wmDeleteWindow, True);

	mb->wnd = XCreateSimpleWindow(mb->dpy, DefaultRootWindow(mb->dpy),
				      0, 0, 100, 100, 0, mb->black, mb->black);

	if (!mb->wnd)
		goto failwnd;

	XStoreName(mb->dpy, mb->wnd, title);
	XSetWMProtocols(mb->dpy, mb->wnd, &mb->wmDelete, 1);
	XSelectInput(mb->dpy, mb->wnd,
		     ExposureMask | StructureNotifyMask |
		     KeyReleaseMask | PointerMotionMask |
		     ButtonPressMask | ButtonReleaseMask);

	mb->gc = XCreateGC(mb->dpy, mb->wnd, 0, 0);

	if (!mb->gc)
		goto fail;

	XSetForeground(mb->dpy, mb->gc, mb->white);
	XSetBackground(mb->dpy, mb->gc, mb->black);
	return 1;
fail:
	XDestroyWindow(mb->dpy, mb->wnd);
failwnd:
	XCloseDisplay(mb->dpy);
	return 0;
}

static void messagebox_cleanup(struct messagebox *mb)
{
	XFreeGC(mb->dpy, mb->gc);
	XDestroyWindow(mb->dpy, mb->wnd);
	XCloseDisplay(mb->dpy);
}

static void messagebox_draw(struct messagebox *mb, const char *text,
			    size_t height)
{
	const char *temp = text, *end;
	size_t i = 0;

	XClearWindow(mb->dpy, mb->wnd);

	while (temp != NULL) {
		end = strchr(temp, '\n');

		XDrawString(mb->dpy, mb->wnd, mb->gc, 10, height + i + 10,
			    temp, end ? (size_t)(end - temp) : strlen(temp));

		temp = end ? (end + 1) : NULL;
		i += height;
	}

	messagebox_draw_button(mb, &mb->ok);
	XFlush(mb->dpy);
}

static void messagebox_show(struct messagebox *mb, int X, int Y, int W, int H)
{
	XSizeHints hints;

	XMoveResizeWindow(mb->dpy, mb->wnd, X, Y, W, H);

	hints.flags = PSize | PMinSize | PMaxSize;
	hints.min_width = hints.max_width = hints.base_width = W;
	hints.min_height = hints.max_height = hints.base_height = H;

	XSetWMNormalHints(mb->dpy, mb->wnd, &hints);
	XMapRaised(mb->dpy, mb->wnd);
	XFlush(mb->dpy);
}

/**************************************************************************
 * A small and simple function that creates a message box with an OK      *
 * button, using ONLY Xlib. It does not return until the user closes the  *
 * message box, using the OK button, the escape key, or the close button. *
 *                                                                        *
 * title: The title of the message box.                                   *
 * text:  The contents of the message box. Use '\n' as a line terminator. *
 **************************************************************************/
void MessageBoxX11(const char *title, const char *text)
{
	int height = 0, direction, ascent, descent, X, Y, W = 0, H;
	const char *end, *temp;
	struct messagebox mb;
	XCharStruct overall;
	size_t lines = 0;
	XFontStruct *font;
	XEvent e;

	if (!messagebox_init(&mb, title))
		return;

	/* Compute the printed width and height of the text */
	if (!(font = XQueryFont(mb.dpy, XGContextFromGC(mb.gc))))
		goto out;

	temp = text;

	while (temp != NULL) {
		end = strchr(temp, '\n');

		XTextExtents(font, temp,
			     end ? (size_t)(end - temp) : strlen(temp),
			     &direction, &ascent, &descent, &overall);

		W = overall.width > W ? overall.width : W;
		height = (ascent + descent) > height ?
			(ascent + descent) : height;
		temp = end ? (end + 1) : NULL;
		++lines;
	}

	/* Compute the shape of the window and adjust the window accordingly */
	W += 20;
	H = lines * height + height + 40;
	X = DisplayWidth(mb.dpy, DefaultScreen(mb.dpy)) / 2 - W / 2;
	Y = DisplayHeight(mb.dpy, DefaultScreen(mb.dpy)) / 2 - H / 2;

	messagebox_show(&mb, X, Y, W, H);

	/* Compute the shape of the OK button */
	XTextExtents(font, "OK", 2, &direction, &ascent, &descent, &overall);

	mb.ok.width = overall.width + 30;
	mb.ok.height = ascent + descent + 5;
	mb.ok.x = W / 2 - mb.ok.width / 2;
	mb.ok.y = H - height - 15;
	mb.ok.textx = mb.ok.x + 15;
	mb.ok.texty = mb.ok.y + mb.ok.height - 3;
	mb.ok.flags = 0;
	mb.ok.text = "OK";

	XFreeFontInfo(NULL, font, 1);

	/* Event loop */
	for (;;) {
		XNextEvent(mb.dpy, &e);
		mb.ok.flags &= ~BUTTON_CLICKED;

		switch (e.type) {
		case MotionNotify:
			if (is_point_inside(&mb.ok, e.xmotion.x, e.xmotion.y)) {
				if (mb.ok.flags & BUTTON_HOVER)
					break;
			} else if (!(mb.ok.flags & BUTTON_HOVER)) {
				break;
			}
			mb.ok.flags ^= BUTTON_HOVER;
			messagebox_draw(&mb, text, height);
			break;
		case ButtonPress:
		case ButtonRelease:
			if (e.xbutton.button != Button1)
				break;
			if (!(mb.ok.flags & BUTTON_HOVER))
				break;

			mb.ok.flags = e.type == ButtonPress ?
				(mb.ok.flags | BUTTON_CLICKED) :
				(mb.ok.flags & ~BUTTON_CLICKED);

			if (!(mb.ok.flags & BUTTON_CLICKED))
				goto out;
			/* fall-through */
		case Expose:
		case MapNotify:
			messagebox_draw(&mb, text, height);
			break;
		case KeyRelease:
			if (XLookupKeysym(&e.xkey, 0) == XK_Escape)
				goto out;
			break;
		case ClientMessage:
			if (e.xclient.data.l[0] == (long)mb.wmDelete)
				goto out;
			break;
		}
	}
out:
	messagebox_cleanup(&mb);
}


int main(void)
{
	MessageBoxX11("Test", "Foo bar test text bla bla\nMulti line\nText");
	return EXIT_SUCCESS;
}
